from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class SetMaxBurnAmountPerMessageParamsJSON(typing.TypedDict):
    burn_limit_per_message: int


@dataclass
class SetMaxBurnAmountPerMessageParams:
    layout: typing.ClassVar = borsh.CStruct("burn_limit_per_message" / borsh.U64)
    burn_limit_per_message: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "SetMaxBurnAmountPerMessageParams":
        return cls(burn_limit_per_message=obj.burn_limit_per_message)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"burn_limit_per_message": self.burn_limit_per_message}

    def to_json(self) -> SetMaxBurnAmountPerMessageParamsJSON:
        return {"burn_limit_per_message": self.burn_limit_per_message}

    @classmethod
    def from_json(
        cls, obj: SetMaxBurnAmountPerMessageParamsJSON
    ) -> "SetMaxBurnAmountPerMessageParams":
        return cls(burn_limit_per_message=obj["burn_limit_per_message"])



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class PlaceOrderParamsJSON(typing.TypedDict):
    source_sell_amount: int
    min_sell_amount: int
    dst_token_mint: list[int]
    dst_buy_amount: int
    eid: int
    target_address: list[int]
    bond_asset_mint: str
    bond_amount: int
    bond_fee: int


@dataclass
class PlaceOrderParams:
    layout: typing.ClassVar = borsh.CStruct(
        "source_sell_amount" / borsh.U64,
        "min_sell_amount" / borsh.U64,
        "dst_token_mint" / borsh.U8[32],
        "dst_buy_amount" / borsh.U64,
        "eid" / borsh.U32,
        "target_address" / borsh.U8[32],
        "bond_asset_mint" / BorshPubkey,
        "bond_amount" / borsh.U64,
        "bond_fee" / borsh.U16,
    )
    source_sell_amount: int
    min_sell_amount: int
    dst_token_mint: list[int]
    dst_buy_amount: int
    eid: int
    target_address: list[int]
    bond_asset_mint: Pubkey
    bond_amount: int
    bond_fee: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "PlaceOrderParams":
        return cls(
            source_sell_amount=obj.source_sell_amount,
            min_sell_amount=obj.min_sell_amount,
            dst_token_mint=obj.dst_token_mint,
            dst_buy_amount=obj.dst_buy_amount,
            eid=obj.eid,
            target_address=obj.target_address,
            bond_asset_mint=obj.bond_asset_mint,
            bond_amount=obj.bond_amount,
            bond_fee=obj.bond_fee,
        )

    def to_encodable(self) -> dict[str, typing.Any]:
        return {
            "source_sell_amount": self.source_sell_amount,
            "min_sell_amount": self.min_sell_amount,
            "dst_token_mint": self.dst_token_mint,
            "dst_buy_amount": self.dst_buy_amount,
            "eid": self.eid,
            "target_address": self.target_address,
            "bond_asset_mint": self.bond_asset_mint,
            "bond_amount": self.bond_amount,
            "bond_fee": self.bond_fee,
        }

    def to_json(self) -> PlaceOrderParamsJSON:
        return {
            "source_sell_amount": self.source_sell_amount,
            "min_sell_amount": self.min_sell_amount,
            "dst_token_mint": self.dst_token_mint,
            "dst_buy_amount": self.dst_buy_amount,
            "eid": self.eid,
            "target_address": self.target_address,
            "bond_asset_mint": str(self.bond_asset_mint),
            "bond_amount": self.bond_amount,
            "bond_fee": self.bond_fee,
        }

    @classmethod
    def from_json(cls, obj: PlaceOrderParamsJSON) -> "PlaceOrderParams":
        return cls(
            source_sell_amount=obj["source_sell_amount"],
            min_sell_amount=obj["min_sell_amount"],
            dst_token_mint=obj["dst_token_mint"],
            dst_buy_amount=obj["dst_buy_amount"],
            eid=obj["eid"],
            target_address=obj["target_address"],
            bond_asset_mint=Pubkey.from_string(obj["bond_asset_mint"]),
            bond_amount=obj["bond_amount"],
            bond_fee=obj["bond_fee"],
        )



import datetime
import logging
from pathlib import Path
import sys
from typing import Any, MutableMapping, Optional


try:
    sys.stdout.reconfigure(line_buffering=False)  # type: ignore
except Exception:
    pass


Logger = logging.Logger | logging.LoggerAdapter


# Simple adapter that supports adding a single string-formattable object as context
class LogContextAdapter(logging.LoggerAdapter):
    __slots__ = ("context",)

    def __init__(self, logger: Logger, context: Any):
        super().__init__(logger)
        self.context = context

    def process(self, msg, kwargs) -> tuple[Any, MutableMapping[str, Any]]:
        return (f"[{self.context}] {msg}", kwargs)


def make_logger(
    name: str,
    stdout: Optional[int],
    file: Optional[Path] = None,
    format: str = "%(asctime)s %(message)s",
) -> logging.Logger:
    logger = logging.getLogger(name)

    # Prevent duplicate initializations
    if logger.hasHandlers() and logger.handlers:
        return logger

    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter(format)

    if stdout:
        handler: logging.Handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(stdout)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    if file:
        handler = logging.FileHandler(file)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)

    # Prevents logging from propagating to the root logger, which would cause logs to appear in stdout if basicConfig is called, even though we're not adding stdout handler
    logger.propagate = False

    logger.debug(f"START {name} {datetime.datetime.today()}")

    return logger


def close_logger(logger: logging.Logger) -> None:
    # Ensures that file handles are released
    for handler in logger.handlers:
        handler.close()
        logger.removeHandler(handler)



from __future__ import annotations
import asyncio
import dataclasses
import typing
from typing import Any

from tronpy.async_tron import (
    AsyncTransaction,
    AsyncTransactionRet,
    AsyncTransactionBuilder,
    AsyncTron,
)
from tronpy.async_contract import AsyncContractMethod
import trontxsize

from ..account import TronAccount
from ..chain import TronChain
from ..chain_client import TronClient
from .transaction import SentTransaction, Transaction


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class TronSentTransaction[Chain: TronChain](SentTransaction[Chain]):
    chain_: Chain
    native: AsyncTransactionRet

    @property
    @typing.override
    def id(self) -> str:
        return self.native.txid

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.chain_

    @typing.override
    async def wait_for_receipt(self, **kwargs: Any) -> dict:
        return await self.native.wait(**kwargs)


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class TronTransaction[Chain: TronChain](Transaction[Chain]):
    chain_: Chain
    native: AsyncTransaction

    @staticmethod
    async def _estimate_energy(
        client: AsyncTron,
        owner: TronAccount[Chain],
        contract_method: AsyncContractMethod,
        *args: Any,
        **kwargs: Any,
    ) -> int:
        assert (contract_address := contract_method._contract.contract_address)

        return await client.get_estimated_energy(
            owner_address=owner.address,
            contract_address=contract_address,
            function_selector=contract_method.function_signature,
            parameter=contract_method._prepare_parameter(*args, **kwargs),
        )

    @staticmethod
    def _extract_price_from_response(response: dict) -> int:
        prices = response.get("prices", "").split(",")
        recent_price: str = prices[-1]
        _timestamp, price = recent_price.split(":")

        return int(price)

    @classmethod
    async def _get_energy_price(cls, client: AsyncTron) -> int:
        response = await client.provider.make_request("wallet/getenergyprices")
        return cls._extract_price_from_response(response)

    @classmethod
    async def _get_bandwidth_price(cls, client: AsyncTron) -> int:
        response = await client.provider.make_request("wallet/getbandwidthprices")
        return cls._extract_price_from_response(response)

    @classmethod
    async def from_builder(
        cls,
        client: TronClient[Chain],
        builder: AsyncTransactionBuilder,
        signer: TronAccount[Chain],
        energy_limit: int,
    ) -> TronTransaction[Chain]:
        transaction: AsyncTransaction = await builder.with_owner(signer.address).build()
        transaction = transaction.sign(signer.private_key_)

        bandwidth_estimate = trontxsize.get_tx_size(transaction.to_json())
        bandwidth_price = await cls._get_bandwidth_price(client.native)

        # Warning: We're overwriting the default fee limit set by the client config
        builder = builder.fee_limit(energy_limit + bandwidth_estimate * bandwidth_price)

        transaction = await builder.build()
        transaction = transaction.sign(signer.private_key_)

        return cls(chain_=client.chain, native=transaction)

    @classmethod
    async def from_contract_method(
        cls,
        client: TronClient[Chain],
        signer: TronAccount[Chain],
        contract_method: AsyncContractMethod,
        *args: Any,
        **kwargs: Any,
    ) -> TronTransaction[Chain]:
        builder: AsyncTransactionBuilder = await contract_method(*args, **kwargs)
        transaction: AsyncTransaction = await builder.with_owner(signer.address).build()
        transaction = transaction.sign(signer.private_key_)

        energy_estimate, energy_price = await asyncio.gather(
            cls._estimate_energy(
                client.native, signer, contract_method, *args, **kwargs
            ),
            cls._get_energy_price(client.native),
        )

        return await cls.from_builder(
            client, builder, signer, energy_estimate * energy_price
        )

    @typing.override
    async def broadcast(self) -> TronSentTransaction[Chain]:
        broadcasted = await self.native.broadcast()
        return TronSentTransaction(chain_=self.chain, native=broadcasted)

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.chain_



from decimal import Decimal
import pprint
from typing import Any

from aiohttp import ClientResponse

from ..asset import Token
from ..chain import Chain


async def check_response(response: ClientResponse) -> Any:
    result = None

    if response.status != 200:
        try:
            result = await response.json()
        except Exception:
            result = await response.text()

        raise RuntimeError(f"API request failed:\n{response}{pprint.pformat(result)}")

    return result


async def to_json(response: ClientResponse) -> Any:
    await check_response(response)
    return await response.json()


async def to_bytes(response: ClientResponse) -> bytes:
    await check_response(response)
    return await response.read()


def chain_balances_in_coins(raw_balances: dict[Token, int]) -> dict[Token, Decimal]:
    return {token: token.to_coins(balance) for token, balance in raw_balances.items()}


def balances_in_coins(
    raw_balances: dict[Chain, dict[Token, int]],
) -> dict[Chain, dict[Token, Decimal]]:
    return {
        chain: chain_balances_in_coins(chain_balances)
        for chain, chain_balances in raw_balances.items()
    }



from .book import OrderDirection, OrderExpiration, OrderFunding
from .orders import CCTPRequest, GasResponse, OrderRequest, OrderResponse
from .quotes import OrderData, Quote, QuoteRequest, LiquiditySource
from .types import Chain as MachChain


__all__ = [
    "CCTPRequest",
    "GasResponse",
    "LiquiditySource",
    "MachChain",
    "OrderData",
    "OrderDirection",
    "OrderExpiration",
    "OrderFunding",
    "OrderRequest",
    "OrderResponse",
    "Quote",
    "QuoteRequest",
]



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class RegisterTristeroOappArgs(typing.TypedDict):
    params: types.register_tristero_o_app_params.RegisterTristeroOAppParams


layout = borsh.CStruct(
    "params" / types.register_tristero_o_app_params.RegisterTristeroOAppParams.layout
)


class RegisterTristeroOappAccounts(typing.TypedDict):
    payer: Pubkey
    oapp: Pubkey
    admin_panel: Pubkey
    oapp_registry: Pubkey
    event_authority: Pubkey
    endpoint_program: Pubkey


def register_tristero_oapp(
    args: RegisterTristeroOappArgs,
    accounts: RegisterTristeroOappAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["payer"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["oapp"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["admin_panel"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["oapp_registry"], is_signer=False, is_writable=True
        ),
        AccountMeta(
            pubkey=accounts["event_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(
            pubkey=accounts["endpoint_program"], is_signer=False, is_writable=False
        ),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"E\r-\xa6\xa5\xb4\n>"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from .ethereum import EthereumTransaction, EthereumSentTransaction
from .transaction import SentTransaction, Transaction
from .solana import SolanaTransaction, SolanaSentTransaction
from .tron import TronTransaction, TronSentTransaction


__all__ = [
    "EthereumSentTransaction",
    "EthereumTransaction",
    "SentTransaction",
    "SolanaTransaction",
    "SolanaSentTransaction",
    "Transaction",
    "TronSentTransaction",
    "TronTransaction",
]



import abc
from abc import ABC
import asyncio
import dataclasses
from decimal import Decimal
import typing
from typing import Optional

from ..asset import Token
from ..client.asset_server import AssetServer
from ..log import LogContextAdapter, Logger
from .types import Quote


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class RiskAnalysis:
    reject: bool


class RiskManager(ABC):
    __slots__ = ("logger",)

    def __init__(self, logger: Logger):
        self.logger = logger

    @abc.abstractmethod
    async def __call__(
        self, src_token: Token, dest_token: Token, quote: Quote
    ) -> RiskAnalysis:
        pass


class BaseSlippageManager(RiskManager):
    @dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
    class RiskAnalysis(RiskAnalysis):
        slippage: Optional[Decimal]
        slippage_tolerance: Decimal

    __slots__ = ("slippage_tolerance",)

    def __init__(
        self,
        slippage_tolerance: Decimal,
        logger: Logger,
    ):
        super().__init__(LogContextAdapter(logger, "Slippage Manager"))
        assert -1.0 <= slippage_tolerance <= 0.0
        self.slippage_tolerance = slippage_tolerance

    @abc.abstractmethod
    def will_check(self, src_token: Token, dest_token: Token) -> bool:
        pass

    @abc.abstractmethod
    def reject_by_default(self) -> bool:
        pass

    @abc.abstractmethod
    async def get_value(self, token: Token, amount: int) -> Decimal:
        pass

    @typing.override
    async def __call__(
        self, src_token: Token, dest_token: Token, quote: Quote
    ) -> RiskAnalysis:
        if not self.will_check(src_token, dest_token):
            self.logger.warning(f"Not checking {src_token} and {dest_token}")

            return self.RiskAnalysis(
                reject=self.reject_by_default(),
                slippage=None,
                slippage_tolerance=self.slippage_tolerance,
            )

        src_value, dest_value = await asyncio.gather(
            self.get_value(src_token, quote.src_amount),
            self.get_value(dest_token, quote.dst_amount),
        )

        slippage = dest_value / src_value - Decimal(1.0)
        self.logger.info(f"{src_token} => {dest_token} slippage: {100 * slippage}%")

        return self.RiskAnalysis(
            reject=slippage < self.slippage_tolerance,
            slippage=slippage,
            slippage_tolerance=self.slippage_tolerance,
        )


class SlippageManager(BaseSlippageManager):
    __slots__ = ("asset_server",)

    def __init__(
        self,
        asset_server: AssetServer,
        slippage_tolerance: Decimal,
        logger: Logger,
    ):
        super().__init__(slippage_tolerance, logger)
        self.asset_server = asset_server

    @typing.override
    def will_check(self, src_token: Token, dest_token: Token) -> bool:
        return self.asset_server.is_supported(
            src_token
        ) and self.asset_server.is_supported(dest_token)

    @typing.override
    def reject_by_default(self) -> bool:
        return True

    @typing.override
    async def get_value(self, token: Token, amount: int) -> Decimal:
        usd_price = await self.asset_server.get_price(token)
        return token.to_coins(amount) * usd_price


# Only checks "similar tokens". Doesn't need to make a network request.
class SimilarTokenSlippageManager(BaseSlippageManager):
    __slots__ = tuple()

    def __init__(self, slippage_tolerance: Decimal, logger: Logger):
        super().__init__(slippage_tolerance, logger)

    @typing.override
    def will_check(self, src_token: Token, dest_token: Token) -> bool:
        return (
            src_token.symbol == dest_token.symbol
            or (src_token.is_usd_stablecoin() and dest_token.is_usd_stablecoin())
            or (src_token.is_eth() and dest_token.is_eth())
            or (src_token.is_btc() and dest_token.is_btc())
            or (src_token.is_eur_stablecoin() and dest_token.is_eur_stablecoin())
            or (src_token.is_gbp_stablecoin() and dest_token.is_gbp_stablecoin())
            or (src_token.is_jpy_stablecoin() and dest_token.is_jpy_stablecoin())
            or (src_token.is_chf_stablecoin() and dest_token.is_chf_stablecoin())
        )

    @typing.override
    def reject_by_default(self) -> bool:
        return False

    @typing.override
    async def get_value(self, token: Token, amount: int) -> Decimal:
        return token.to_coins(amount)



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class OrderJSON(typing.TypedDict):
    order_id: int
    user_pubkey: str
    user_token_addr: str
    source_token_mint: str
    match_pubkey: typing.Optional[str]
    dst_token_mint: list[int]
    source_sell_amount: int
    dst_buy_amount: int
    min_sell_amount: int
    settled: int
    eid: int
    bump: int
    is_valiable: bool
    target_address: list[int]
    bond_asset_mint: str
    bond_amount: int
    bond_fee: int


@dataclass
class Order:
    discriminator: typing.ClassVar = b"\x86\xad\xdf\xb9MV\x1c3"
    layout: typing.ClassVar = borsh.CStruct(
        "order_id" / borsh.U64,
        "user_pubkey" / BorshPubkey,
        "user_token_addr" / BorshPubkey,
        "source_token_mint" / BorshPubkey,
        "match_pubkey" / borsh.Option(BorshPubkey),
        "dst_token_mint" / borsh.U8[32],
        "source_sell_amount" / borsh.U64,
        "dst_buy_amount" / borsh.U64,
        "min_sell_amount" / borsh.U64,
        "settled" / borsh.U64,
        "eid" / borsh.U32,
        "bump" / borsh.U8,
        "is_valiable" / borsh.Bool,
        "target_address" / borsh.U8[32],
        "bond_asset_mint" / BorshPubkey,
        "bond_amount" / borsh.U64,
        "bond_fee" / borsh.U16,
    )
    order_id: int
    user_pubkey: Pubkey
    user_token_addr: Pubkey
    source_token_mint: Pubkey
    match_pubkey: typing.Optional[Pubkey]
    dst_token_mint: list[int]
    source_sell_amount: int
    dst_buy_amount: int
    min_sell_amount: int
    settled: int
    eid: int
    bump: int
    is_valiable: bool
    target_address: list[int]
    bond_asset_mint: Pubkey
    bond_amount: int
    bond_fee: int

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Order"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Order"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Order"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Order":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Order.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            order_id=dec.order_id,
            user_pubkey=dec.user_pubkey,
            user_token_addr=dec.user_token_addr,
            source_token_mint=dec.source_token_mint,
            match_pubkey=dec.match_pubkey,
            dst_token_mint=dec.dst_token_mint,
            source_sell_amount=dec.source_sell_amount,
            dst_buy_amount=dec.dst_buy_amount,
            min_sell_amount=dec.min_sell_amount,
            settled=dec.settled,
            eid=dec.eid,
            bump=dec.bump,
            is_valiable=dec.is_valiable,
            target_address=dec.target_address,
            bond_asset_mint=dec.bond_asset_mint,
            bond_amount=dec.bond_amount,
            bond_fee=dec.bond_fee,
        )

    def to_json(self) -> OrderJSON:
        return {
            "order_id": self.order_id,
            "user_pubkey": str(self.user_pubkey),
            "user_token_addr": str(self.user_token_addr),
            "source_token_mint": str(self.source_token_mint),
            "match_pubkey": (
                None if self.match_pubkey is None else str(self.match_pubkey)
            ),
            "dst_token_mint": self.dst_token_mint,
            "source_sell_amount": self.source_sell_amount,
            "dst_buy_amount": self.dst_buy_amount,
            "min_sell_amount": self.min_sell_amount,
            "settled": self.settled,
            "eid": self.eid,
            "bump": self.bump,
            "is_valiable": self.is_valiable,
            "target_address": self.target_address,
            "bond_asset_mint": str(self.bond_asset_mint),
            "bond_amount": self.bond_amount,
            "bond_fee": self.bond_fee,
        }

    @classmethod
    def from_json(cls, obj: OrderJSON) -> "Order":
        return cls(
            order_id=obj["order_id"],
            user_pubkey=Pubkey.from_string(obj["user_pubkey"]),
            user_token_addr=Pubkey.from_string(obj["user_token_addr"]),
            source_token_mint=Pubkey.from_string(obj["source_token_mint"]),
            match_pubkey=(
                None
                if obj["match_pubkey"] is None
                else Pubkey.from_string(obj["match_pubkey"])
            ),
            dst_token_mint=obj["dst_token_mint"],
            source_sell_amount=obj["source_sell_amount"],
            dst_buy_amount=obj["dst_buy_amount"],
            min_sell_amount=obj["min_sell_amount"],
            settled=obj["settled"],
            eid=obj["eid"],
            bump=obj["bump"],
            is_valiable=obj["is_valiable"],
            target_address=obj["target_address"],
            bond_asset_mint=Pubkey.from_string(obj["bond_asset_mint"]),
            bond_amount=obj["bond_amount"],
            bond_fee=obj["bond_fee"],
        )



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.system_program import ID as SYS_PROGRAM_ID
from spl.token.constants import TOKEN_PROGRAM_ID
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class StartChallengeArgs(typing.TypedDict):
    params: types.challenge_params.ChallengeParams


layout = borsh.CStruct("params" / types.challenge_params.ChallengeParams.layout)


class StartChallengeAccounts(typing.TypedDict):
    authority: Pubkey
    trade_match: Pubkey


def start_challenge(
    args: StartChallengeArgs,
    accounts: StartChallengeAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["authority"], is_signer=True, is_writable=True),
        AccountMeta(pubkey=accounts["trade_match"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=SYS_PROGRAM_ID, is_signer=False, is_writable=False),
        AccountMeta(pubkey=TOKEN_PROGRAM_ID, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xf1`\xfd\xbb\xb1\x9e\x10z"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing

from ..asset import Token
from ..transaction import TronSentTransaction
from .scanner import Scanner


class TronScan(Scanner):
    __slots__ = tuple()

    url = "https://tronscan.io/#"

    @property
    @typing.override
    def base_url(self) -> str:
        return self.url

    @typing.override
    def address(self, address: str) -> str:
        return f"{self.url}/address/{address}"

    @typing.override
    def transaction(self, transaction: TronSentTransaction) -> str:
        return f"{self.url}/transaction/{transaction.id}"

    @typing.override
    def token(self, token: Token) -> str:
        return f"{self.url}/token20/{token.address}"



from __future__ import annotations
from decimal import Decimal
from importlib import resources
from importlib.resources.abc import Traversable
import json
import os
from pathlib import Path
from typing import Any, Annotated, Optional

from anchorpy_core.idl import Idl
from pydantic import BaseModel, PlainValidator
import yaml

from .account import AccountManager, SimpleAccountManager
from .chain import Chain, constants


class AddURLMixin(BaseModel):
    def add_url[Self: AddURLMixin](self: Self, url: str) -> Self:
        return type(self)(
            **{key: f"{url}{value}" for key, value in self.model_dump().items()}
        )


class BackendAPIRoutes(AddURLMixin):
    orders: str = "/v1/orders"
    cctp_orders: str = "/v1/orders/cctp"
    order_status: str = "/v1/orders/status"
    gas: str = "/v1/orders/gas"
    quotes: str = "/v1/quotes"
    points: str = "/v1/points"
    token_balances: str = "/tokenBalances"
    get_config: str = "/get_config"


class BackendConfig(BaseModel):
    url: str
    endpoints: BackendAPIRoutes = BackendAPIRoutes()


class MilesServerV2APIRoutes(AddURLMixin):
    leaderboard: str = "/v1/leaderboard"
    breakdown: str = "/v1/breakdown"
    miles: str = "/v1/miles"
    bound_referral_code: str = "/v1/referral/bound-code"
    all_bound_referral_codes: str = "/v1/referral/bound_referral_codes"
    generated_referral_code: str = "/v1/referral/generated-code"
    all_generated_referral_codes: str = "/v1/referral/generated_codes"
    generate_referral_code: str = "/v1/referral/generate-code"


class MilesServerV2Config(BaseModel):
    url: str
    endpoints: MilesServerV2APIRoutes = MilesServerV2APIRoutes()


class TokenServerAPIRoutes(AddURLMixin):
    assets: str = "/v1/assets"
    prices: str = "/v1/prices"
    users: str = "/v1/users"


class TokenServerConfig(BaseModel):
    url: str
    endpoints: TokenServerAPIRoutes = TokenServerAPIRoutes()


class TradingConfig(BaseModel):
    slippage_tolerance: Decimal


def parse_endpoint_uris(data: dict[str, str]) -> dict[Chain, str]:
    return {
        Chain.from_id(*chain_id.split(":")): endpoint_uri
        for chain_id, endpoint_uri in data.items()
        if chain_id in constants.CHAIN_ID_TO_CHAIN
    }


class Config(BaseModel):
    backend: BackendConfig
    miles_server: MilesServerV2Config
    token_server: TokenServerConfig
    trading: TradingConfig
    endpoint_uris: Annotated[dict[Chain, str], PlainValidator(parse_endpoint_uris)]


def load_abi(path: Traversable, raw: bool = False) -> Any:
    with path.open("r") as abi:
        if raw:
            return abi.read()
        else:
            return json.load(abi)


config_path = Path(os.environ.get("CONFIG_PATH", "config.yaml"))
with open(config_path) as config_file:
    full_config = yaml.safe_load(config_file)

config = Config.model_validate(full_config["mach_client"])

accounts: Optional[AccountManager] = (
    None
    if "accounts" not in full_config
    else SimpleAccountManager.from_key_values(**full_config["accounts"])
)

abi_path = resources.files("abi")

ethereum_order_book_abi = load_abi(abi_path / "ethereum" / "order_book.json")
tron_order_book_abi = load_abi(abi_path / "tron" / "order_book.json")
tristero_idl = Idl.from_json(load_abi(abi_path / "solana" / "tristero.json", raw=True))

ethereum_cctp_token_messenger_abi = load_abi(
    abi_path / "ethereum" / "cctp_token_messenger.json"
)

erc20_abi = load_abi(abi_path / "ethereum" / "erc20.json")
trc20_abi = load_abi(abi_path / "tron" / "trc20.json")



from __future__ import annotations
import abc
from abc import ABC
import dataclasses
import typing
from typing import Optional

from ..chain import Chain, EthereumChain, SolanaChain, TronChain
from .account import Account, AccountBase, AccountID
from .hdwallet import Wallet


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class AccountIDManager(ABC):
    @abc.abstractmethod
    def get(self, chain: Chain) -> Optional[AccountID]:
        pass

    @abc.abstractmethod
    def __getitem__(self, chain: Chain) -> AccountID:
        pass

    def __contains__(self, chain: Chain) -> bool:
        return self.get(chain) is not None


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class AccountIDManagerAdapter(AccountIDManager):
    account_manager: AccountManager

    @typing.override
    def get(self, chain: Chain) -> Optional[AccountID]:
        if not (account := self.account_manager.get(chain)):
            return None

        return account.downcast()

    @typing.override
    def __getitem__(self, chain: Chain) -> AccountID:
        return self.account_manager[chain].downcast()


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class AccountManager(ABC):
    @abc.abstractmethod
    def get(self, chain: Chain) -> Optional[Account]:
        pass

    @abc.abstractmethod
    def __getitem__(self, chain: Chain) -> Account:
        pass

    def downcast(self) -> AccountIDManager:
        return AccountIDManagerAdapter(account_manager=self)

    def __contains__(self, chain: Chain) -> bool:
        return self.get(chain) is not None


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class SimpleAccountManager(AccountManager):
    private_keys: dict[type[Chain], Optional[str]]

    @classmethod
    def from_account(cls, account: Account) -> SimpleAccountManager:
        return cls(private_keys={type(account.chain): account.private_key})

    @classmethod
    def from_key_values(
        cls,
        *,
        ethereum: Optional[str] = None,
        solana: Optional[str] = None,
        tron: Optional[str] = None,
    ) -> SimpleAccountManager:
        return cls(
            private_keys={
                EthereumChain: ethereum,
                SolanaChain: solana,
                TronChain: tron,
            }
        )

    @typing.override
    def get(self, chain: Chain) -> Optional[Account]:
        chain_type = type(chain)

        if not (private_key := self.private_keys.get(chain_type)):
            return None

        return Account.from_str(chain, private_key)

    @typing.override
    def __getitem__(self, chain: Chain) -> Account:
        if not (account := self.get(chain)):
            raise KeyError(f"No private key for {chain}")

        return account


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class SimpleAccountIDManager(AccountIDManager):
    addresses: dict[type[Chain], Optional[str]]

    @classmethod
    def from_account(cls, account: AccountBase) -> SimpleAccountIDManager:
        return cls(addresses={type(account.chain): account.address})

    @classmethod
    def from_key_values(
        cls,
        *,
        ethereum: Optional[str] = None,
        solana: Optional[str] = None,
        tron: Optional[str] = None,
    ) -> SimpleAccountIDManager:
        return cls(
            addresses={
                EthereumChain: ethereum,
                SolanaChain: solana,
                TronChain: tron,
            }
        )

    @typing.override
    def get(self, chain: Chain) -> Optional[AccountID]:
        chain_type = type(chain)

        if not (address := self.addresses.get(chain_type)):
            return None

        return AccountID.from_str(chain, address)

    @typing.override
    def __getitem__(self, chain: Chain) -> AccountID:
        if not (account_id := self.get(chain)):
            raise KeyError(f"No address for {chain}")

        return account_id


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class HDWalletAccountManager(AccountManager):
    wallet: Wallet

    @typing.override
    def get(self, chain: Chain) -> Optional[Account]:
        return self[chain]

    @typing.override
    def __getitem__(self, chain: Chain) -> Account:
        return self.wallet.account(chain)



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
import borsh_construct as borsh


class BurnTokenCustodyParamsJSON(typing.TypedDict):
    amount: int


@dataclass
class BurnTokenCustodyParams:
    layout: typing.ClassVar = borsh.CStruct("amount" / borsh.U64)
    amount: int

    @classmethod
    def from_decoded(cls, obj: Container) -> "BurnTokenCustodyParams":
        return cls(amount=obj.amount)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"amount": self.amount}

    def to_json(self) -> BurnTokenCustodyParamsJSON:
        return {"amount": self.amount}

    @classmethod
    def from_json(cls, obj: BurnTokenCustodyParamsJSON) -> "BurnTokenCustodyParams":
        return cls(amount=obj["amount"])



from __future__ import annotations
import abc
from abc import ABC
import dataclasses
import typing
from typing import Any

from hdwallet import HDWallet
from solders.keypair import Keypair

from ..chain import Chain, EthereumChain, SolanaChain, TronChain


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class AccountBase[ChainType: Chain](ABC):
    chain: ChainType

    @property
    @abc.abstractmethod
    def address(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def id(self) -> str:
        pass

    @abc.abstractmethod
    def encode_address(self) -> bytes:
        pass

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.address})"

    def __str__(self):
        return self.address


# CAIP-10 Account ID - acts as proxy for an address on a chain
# https://chainagnostic.org/CAIPs/caip-10
# Purely the address/public key component
@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class AccountID[ChainType: Chain](AccountBase[ChainType]):
    @staticmethod
    def from_str(chain: ChainType, address: str) -> AccountID[ChainType]:
        match chain:
            case EthereumChain():
                return EthereumAccountID.from_address(chain, address)
            case SolanaChain():
                return SolanaAccountID.from_pubkey(chain, address)
            case TronChain():
                return TronAccountID.from_address(chain, address)
            case _:
                raise NotImplementedError(f"Unsupported chain: {chain}")

    @staticmethod
    def from_hdwallet(chain: Chain, hdwallet: HDWallet) -> AccountID:
        return AccountID.from_str(chain, hdwallet.address())

    @staticmethod
    def from_id(namespace: str, reference: str, address: str) -> AccountID:
        chain = Chain.from_id(namespace, reference)
        return AccountID.from_str(chain, address)

    @classmethod
    def from_id_str(cls, id: str) -> AccountID:
        namespace, reference, address = id.split(":")
        return cls.from_id(namespace, reference, address)

    @property
    @typing.override
    def id(self) -> str:
        return f"{self.chain.id}:{self.address}"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, AccountID) and self.id == other.id


# An "owned" address, ie. a keypair or private key
@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class Account[ChainType: Chain](AccountBase[ChainType]):
    @staticmethod
    def from_str(chain: ChainType, private_key: str) -> Account[ChainType]:
        match chain:
            case EthereumChain():
                return EthereumAccount.from_private_key(chain, private_key)
            case SolanaChain():
                return SolanaAccount.from_private_key(chain, private_key)
            case TronChain():
                return TronAccount.from_private_key(chain, private_key)
            case _:
                raise NotImplementedError(f"Unsupported chain: {chain}")

    @staticmethod
    def from_hdwallet(chain: Chain, hdwallet: HDWallet) -> Account:
        private_key = typing.cast(str, hdwallet.private_key())

        match chain:
            case SolanaChain():
                keypair = Keypair.from_seed(bytes.fromhex(private_key))
                return SolanaAccount(chain=chain, keypair=keypair)

            case _:
                return Account.from_str(chain, private_key)

    @staticmethod
    def from_id(namespace: str, reference: str, private_key: str) -> Account:
        chain = Chain.from_id(namespace, reference)
        return Account.from_str(chain, private_key)

    @classmethod
    def from_id_str(cls, id: str) -> Account:
        namespace, reference, private_key = id.split(":")
        return cls.from_id(namespace, reference, private_key)

    @property
    @abc.abstractmethod
    def private_key(self) -> str:
        pass

    @abc.abstractmethod
    def downcast(self) -> AccountID[ChainType]:
        pass

    @property
    @typing.override
    def id(self) -> str:
        return f"{self.chain.id}:{self.private_key}"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Account) and self.id == other.id


# Avoid a circular import
from .ethereum import EthereumAccount, EthereumAccountID  # noqa: E402
from .solana import SolanaAccount, SolanaAccountID  # noqa: E402
from .tron import TronAccount, TronAccountID  # noqa: E402



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class SetTokenControllerParamsJSON(typing.TypedDict):
    token_controller: str


@dataclass
class SetTokenControllerParams:
    layout: typing.ClassVar = borsh.CStruct("token_controller" / BorshPubkey)
    token_controller: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "SetTokenControllerParams":
        return cls(token_controller=obj.token_controller)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"token_controller": self.token_controller}

    def to_json(self) -> SetTokenControllerParamsJSON:
        return {"token_controller": str(self.token_controller)}

    @classmethod
    def from_json(cls, obj: SetTokenControllerParamsJSON) -> "SetTokenControllerParams":
        return cls(token_controller=Pubkey.from_string(obj["token_controller"]))



# Reference:
# https://github.com/circlefin/solana-cctp-contracts/blob/master/examples/utils.ts
# https://github.com/tristeroresearch/cache-half-full/blob/dev-solana-add-bond/backend/solutils.py

import functools
import typing
from typing import Iterable

from solders.pubkey import Pubkey


def encode_address(pubkey: Pubkey) -> bytes:
    return bytes(pubkey)


@functools.cache
def find_program_address_and_bump(
    label: str,
    program_id: Pubkey,
    extra_seeds: Iterable[str | bytes | bytearray | Pubkey] = tuple(),
) -> tuple[Pubkey, int]:
    seeds = [label.encode()]

    for seed in extra_seeds:
        match seed:
            case str():
                seeds.append(seed.encode())
            case bytes() | bytearray():
                seeds.append(seed)
            case Pubkey() | memoryview():
                seeds.append(bytes(seed))
            case _ as arg:
                typing.assert_never(arg)

    return Pubkey.find_program_address(seeds, program_id)


def find_program_address(
    label: str,
    program_id: Pubkey,
    extra_seeds: Iterable[str | bytes | bytearray | Pubkey] = tuple(),
) -> Pubkey:
    return find_program_address_and_bump(label, program_id, extra_seeds)[0]



from .chain_client import ChainClient
from .ethereum import EthereumClient
from .solana import SolanaClient
from .tron import TronClient


__all__ = [
    "ChainClient",
    "EthereumClient",
    "SolanaClient",
    "TronClient",
]



from .account import Account, AccountID, AccountIDManager, AccountManager, Wallet
from .asset import Asset, ApprovableToken, NativeCoin, Token
from .chain import Chain
from .chain.constants import SupportedChain
from .chain_client import ChainClient
from .client import AssetServer, MachClient, RiskManager
from .scanner import Scanner
from .transaction import SentTransaction, Transaction
from .log import LogContextAdapter, Logger


__all__ = [
    "Account",
    "AccountID",
    "AccountIDManager",
    "AccountManager",
    "Asset",
    "AssetServer",
    "ApprovableToken",
    "Chain",
    "ChainClient",
    "LogContextAdapter",
    "Logger",
    "MachClient",
    "NativeCoin",
    "RiskManager",
    "Scanner",
    "SentTransaction",
    "SupportedChain",
    "Token",
    "Transaction",
    "Wallet"
]



from .asset import Asset, GenericAsset
from .ethereum import EthereumToken
from .solana import SolanaToken
from .token import ApprovableToken, NativeCoin, Token
from .tron import TronToken


__all__ = [
    "ApprovableToken",
    "Asset",
    "GenericAsset",
    "SolanaToken",
    "Token",
    "EthereumToken",
    "NativeCoin",
    "TronToken",
]



from __future__ import annotations
import asyncio
from datetime import datetime, timedelta
import logging
import pprint
import typing
from typing import (
    Any,
    AsyncGenerator,
    Coroutine,
    Optional,
)

from aiohttp import ClientSession, WSMsgType
import cachebox
from cachebox import Cache
from pydantic import TypeAdapter

from .. import config, log
from ..account import Account, AccountBase, AccountID, SolanaAccountID
from ..asset import ApprovableToken, NativeCoin, Token
from ..chain import Chain, constants as chain_constants
from ..chain_client import ChainClient
from ..log import LogContextAdapter, Logger
from ..utility import solana
from ..transaction import SentTransaction, Transaction
from . import cctp, order_book, utility
from .asset_server import AssetServer
from .event import (
    ApprovalFailed,
    BroadcastTransactionFailed,
    CreateTransactionFailed,
    DestinationNotReceived,
    InsufficientSourceBalance,
    OrderRejected,
    TradeEvent,
    QuoteFailed,
    QuoteLiquidityUnavailable,
    RiskManagerRejection,
    SourceNotWithdrawn,
    SubmitOrderFailed,
    Trade,
    TransactionError,
    WaitForTransactionReceiptFailed,
)
from .miles_server import MilesServerV2Mixin
from .risk_manager import RiskManager, SimilarTokenSlippageManager
from .types import (
    CCTPRequest,
    GasResponse,
    LiquiditySource,
    MachChain,
    OrderRequest,
    OrderResponse,
    Quote,
    QuoteRequest,
)


_logger = log.make_logger("mach-client", logging.INFO)


# This class is made to be used as a singleton (see below)
class MachClient(MilesServerV2Mixin):
    __slots__ = (
        "logger",
        "routes",
        "_session",
        "deployments",
        "chains",
        "tokens",
    )

    @staticmethod
    def _key_maker(_args: tuple, kwargs: dict) -> Any:
        return (
            kwargs.get("backend_url", config.config.backend.url),
            kwargs.get("logger", _logger).name,
        )

    @classmethod
    @cachebox.cached(Cache(0), key_maker=_key_maker, copy_level=2)
    async def create(
        cls,
        *,
        backend_url: str = config.config.backend.url,
        miles_url: str = config.config.miles_server.url,
        logger: Logger = _logger,
    ) -> MachClient:
        client = cls(
            backend_url=backend_url,
            miles_url=miles_url,
            logger=logger,
        )
        await client.init()
        return client

    def __init__(self, *, backend_url: str, miles_url: str, logger: Logger):
        super().__init__(miles_url)
        self.logger = logger
        self.routes = config.config.backend.endpoints.add_url(backend_url)
        self.deployments: dict[Chain, dict[str, Any]] = {}
        self.chains: set[Chain] = set()
        self.tokens: dict[Chain, set[Token]] = {}

    @property
    @typing.override
    def session(self) -> ClientSession:
        return self._session

    async def init(self) -> None:
        self._session = ClientSession()
        self.session.headers.update(
            (
                ("accept", "application/json"),
                ("Content-Type", "application/json"),
            )
        )

        await self.refresh()

    async def close(self) -> None:
        if hasattr(self, "session"):
            await self.session.close()

    async def refresh(self) -> None:
        await self.refresh_config()

    async def _process_token_data(
        self, chain_client: ChainClient, symbol: str, token_info: dict[str, Any]
    ) -> None:
        chain = chain_client.chain

        # This is the native asset, not a token
        if token_info.get("wrapped"):
            return

        self.tokens[chain].add(
            await Token.register(
                chain_client,
                token_info["address"],
                symbol,
                token_info["decimals"],
            )
        )

    async def _process_chain_data(
        self, chain_name: str, chain_data: dict[str, Any]
    ) -> Optional[tuple[Chain, dict[str, Any]]]:
        mach_chain = MachChain(chain_name)
        if not (chain := mach_chain.try_to_chain()):
            self.logger.warning(f"Chain {chain_name} is not supported")
            return

        # See note in CHAIN_TO_LAYERZERO_ID declaration
        layerzero_id = chain_data["lz_cid"]
        chain_constants.CHAIN_TO_LAYERZERO_ID[chain] = layerzero_id
        chain_constants.LAYERZERO_ID_TO_CHAIN[layerzero_id] = chain

        if cctp_domain := chain_data.get("cctp_id"):
            chain_constants.CCTP_DOMAINS[chain] = cctp_domain

        if chain not in config.config.endpoint_uris:
            self.logger.warning(f"{chain_name} endpoint URI missing from config")
            return

        result = (chain, chain_data)

        chain_client = await ChainClient.create(chain)

        self.chains.add(chain)

        if chain not in self.tokens:
            self.tokens[chain] = set()

        await asyncio.gather(
            *(
                self._process_token_data(chain_client, symbol, token_info)
                for symbol, token_info in chain_data["assets"].items()
            )
        )

        # Remove redundant data
        del chain_data["assets"]
        del chain_data["chain_id"]
        del chain_data["lz_cid"]

        return result

    async def refresh_config(self) -> None:
        async with self.session.get(self.routes.get_config) as response:
            raw_deployments: dict[str, dict[str, Any]] = (
                await utility.to_json(response)
            )["deployments"]

        result = await asyncio.gather(
            *[
                self._process_chain_data(chain_name, chain_data)
                for chain_name, chain_data in raw_deployments.items()
            ]
        )

        self.deployments = dict(filter(lambda data: data, result))  # type: ignore

    async def submit_order(
        self,
        place_order_transaction: SentTransaction,
        referral_code: Optional[str] = None,
    ) -> OrderResponse:
        order_request = OrderRequest(
            chain=MachChain.from_chain(place_order_transaction.chain),
            place_taker_tx=place_order_transaction.id,
            referral_code=referral_code,
        )

        async with self.session.post(
            self.routes.orders, json=order_request.model_dump()
        ) as response:
            bytes_result = await utility.to_bytes(response)

        return OrderResponse.model_validate_json(bytes_result)

    async def submit_cctp_order(
        self, burn_transaction: SentTransaction
    ) -> OrderResponse:
        cctp_request = CCTPRequest(
            chain=MachChain.from_chain(burn_transaction.chain),
            burn_tx=burn_transaction.id,
        )

        async with self.session.post(
            self.routes.cctp_orders, json=cctp_request.model_dump()
        ) as response:
            bytes_result = await utility.to_bytes(response)

        return OrderResponse.model_validate_json(bytes_result)

    async def watch_order_status(
        self, order: OrderResponse
    ) -> AsyncGenerator[OrderResponse]:
        url = f"{self.routes.order_status}/{order.id}"

        async with self.session.ws_connect(url) as websocket:
            async for message in websocket:
                if message.type != WSMsgType.TEXT:
                    continue

                yield OrderResponse.model_validate_json(message.data)

    async def poll_order_status(
        self, order: OrderResponse, poll_time: int
    ) -> AsyncGenerator[Optional[OrderResponse]]:
        url = f"{self.routes.order_status}/{order.id}"

        async with self.session.ws_connect(url) as websocket:
            iter = aiter(websocket)

            while True:
                try:
                    message = await asyncio.wait_for(anext(iter), timeout=poll_time)

                except TimeoutError:
                    yield None
                    continue

                except StopAsyncIteration:
                    break

                if message.type != WSMsgType.TEXT:
                    yield None

                yield OrderResponse.model_validate_json(message.data)

    _orders_validator = TypeAdapter(list[OrderResponse])

    async def get_orders(self, account: AccountBase) -> list[OrderResponse]:
        params = {
            "wallet": account.address,
        }

        async with self.session.get(self.routes.orders, params=params) as response:
            bytes_result = await utility.to_bytes(response)

        return self._orders_validator.validate_json(bytes_result)

    async def estimate_gas(self, chain: Chain) -> GasResponse:
        params = {"chain": MachChain.from_chain(chain).name}

        async with self.session.get(self.routes.gas, params=params) as response:
            bytes_result = await utility.to_bytes(response)

        return GasResponse.model_validate_json(bytes_result)

    async def request_quote(
        self,
        src_token: Token,
        dest_token: Token,
        amount: int,
        src_account: AccountBase,
        dest_account: AccountBase,
    ) -> Quote:
        assert NativeCoin not in (
            type(src_token),
            type(dest_token),
        ), "Native coins not supported"
        assert src_token.chain == src_account.chain
        assert dest_token.chain == dest_account.chain

        quote_request = QuoteRequest(
            wallet_address=src_account.address,
            target_address=dest_account.address,
            src_chain=MachChain.from_chain(src_token.chain),
            dst_chain=MachChain.from_chain(dest_token.chain),
            src_asset_address=src_token.address,
            dst_asset_address=dest_token.address,
            src_amount=amount,
        )

        async with self.session.post(
            self.routes.quotes, json=quote_request.model_dump()
        ) as response:
            bytes_result = await utility.to_bytes(response)

        return Quote.model_validate_json(bytes_result)

    def contract_address(self, chain: Chain, name: str) -> str:
        return self.deployments[chain]["contracts"][name]

    def get_contract_address(self, chain: Chain, name: str) -> Optional[str]:
        return self.deployments[chain]["contracts"].get(name)

    def order_book_address(self, chain: Chain) -> str:
        if address := self.get_contract_address(chain, "order_book_v2"):
            return address

        return self.contract_address(chain, "order_book")

    def cctp_message_transmitter_address(self, chain: Chain) -> str:
        return self.contract_address(chain, "cctp_message_transmitter")

    def cctp_token_messenger_minter_address(self, chain: Chain) -> str:
        return self.contract_address(chain, "cctp_token_messenger")

    async def place_logged_transaction(
        self,
        transaction: Transaction,
        logger: Logger,
    ) -> TransactionError | SentTransaction:
        logger.info("Broadcasting transaction")

        try:
            sent_transaction: SentTransaction = await transaction.broadcast()
        except Exception as e:
            logger.error("Failed to broadcast transaction:", exc_info=e)
            return BroadcastTransactionFailed(transaction, e, {})

        logger.info("Waiting for transaction receipt")

        try:
            receipt = await sent_transaction.wait_for_receipt()
        except Exception as e:
            logger.error(f"Could not get receipt for {sent_transaction.id}", exc_info=e)
            return WaitForTransactionReceiptFailed(sent_transaction, e, {})

        logger.info(f"Received receipt: {receipt}")

        return sent_transaction

    async def approve_token[ChainType: Chain](
        self,
        src_token: ApprovableToken[ChainType],
        account: Account[ChainType],
        spender: AccountID[ChainType],
        amount: int,
        logger: Logger,
    ) -> Optional[ApprovalFailed]:
        allowance = await src_token.get_allowance(account.downcast(), spender)

        if allowance >= amount:
            logger.debug(
                f"{spender} has allowance {src_token.format_amount(allowance)} >= required amount {src_token.format_amount(amount)}"
            )
            return

        logger.info(
            f"Placing approval transaction for {spender} to spend {src_token.format_amount(amount)}"
        )

        try:
            receipt = await src_token.approve(
                account,
                spender,
                amount,
            )
        except Exception as e:
            logger.error("Failed to create approval transaction:", exc_info=e)
            return ApprovalFailed(
                src_token,
                amount,
                account.downcast(),
                spender,
                e,
            )

        logger.info(f"Approval granted with receipt: {receipt}")

    async def place_trade[SrcChain: Chain, DestChain: Chain](
        self,
        *,
        src_token: Token[SrcChain],
        dest_token: Token[DestChain],
        amount: int,
        account: Account[SrcChain],
        recipient: AccountID[DestChain],
        risk_manager: Optional[RiskManager] = None,
        referral_code: Optional[str] = None,
        max_wait_time: timedelta = timedelta(minutes=15),
        quote: Optional[Quote] = None,
        logger: Optional[Logger] = None,
    ) -> TradeEvent:
        assert NativeCoin not in (
            type(src_token),
            type(dest_token),
        ), "Native coins not supported"
        assert src_token.chain == account.chain
        assert dest_token.chain == recipient.chain
        assert amount > 0

        if not logger:
            logger = self.logger

        logger = LogContextAdapter(logger, f"{src_token} => {dest_token}")

        if not risk_manager:
            risk_manager = SimilarTokenSlippageManager(
                config.config.trading.slippage_tolerance, logger
            )

        initial_src_balance = await src_token.get_balance(account.downcast())
        logger.debug(f"{initial_src_balance=}")

        if initial_src_balance < amount:
            logger.warning("Insufficient source balance")
            return InsufficientSourceBalance(
                src_token, initial_src_balance, account.downcast()
            )

        if not quote:
            try:
                quote = await self.request_quote(
                    src_token,
                    dest_token,
                    amount,
                    account.downcast(),
                    recipient,
                )
            except Exception as e:
                logger.error("Quote request failed:", exc_info=e)
                return QuoteFailed(
                    (src_token, dest_token), amount, account.downcast(), e
                )

        logger.info("Quote:")
        logger.info(pprint.pformat(quote))

        assert quote.src_amount == amount

        if quote.liquidity_source == LiquiditySource.unavailable:
            return QuoteLiquidityUnavailable(
                (src_token, dest_token), amount, account.downcast(), quote
            )

        elif quote.reject_order:
            return OrderRejected(
                (src_token, dest_token), amount, account.downcast(), quote
            )

        logger.info(f"Filling order through {quote.liquidity_source}")

        if (risk_analysis := await risk_manager(src_token, dest_token, quote)).reject:
            logger.warning("Order rejected by risk manager with analysis:")
            logger.warning(pprint.pformat(risk_analysis))
            return RiskManagerRejection(
                (src_token, dest_token), amount, quote, risk_analysis
            )

        src_client = await ChainClient.create(src_token.chain)

        # Needs to be computed before the order is placed
        initial_dest_balance = await dest_token.get_balance(recipient)

        error_data = {
            "pair": (src_token, dest_token),
            "quote": quote,
        }

        if quote.liquidity_source == LiquiditySource.market_maker:
            error_data.update((("type", "placeOrder"),))

            spender = AccountID.from_str(
                src_token.chain, self.order_book_address(src_token.chain)
            )

            # In the case of Solana, the spender is not the order book program itself but the oapp PDA
            if isinstance(spender, SolanaAccountID):
                spender_pubkey = solana.find_program_address(
                    "TristeroOapp", spender.pubkey
                )
                spender = SolanaAccountID(chain=spender.chain, pubkey=spender_pubkey)

            create_transaction_coro = order_book.create_place_order_transaction(
                src_client=src_client,
                src_token=src_token,
                order_data=quote.order_data,
                account=account,
            )

            def submit_order(
                sent_transaction: SentTransaction[SrcChain],
            ) -> Coroutine[Any, Any, OrderResponse]:
                return self.submit_order(sent_transaction, referral_code)

        else:  # quote.liquidity_source == LiquiditySource.cctp_direct
            error_data.update((("type", "CCTP"),))

            spender = AccountID.from_str(
                src_token.chain,
                self.cctp_token_messenger_minter_address(src_token.chain),
            )

            create_transaction_coro = cctp.create_cctp_burn_transaction(
                client=self,
                src_client=src_client,
                src_token=src_token,
                dest_token=dest_token,
                order_data=quote.order_data,
                account=account,
                recipient=recipient,
            )

            def submit_order(
                sent_transaction: SentTransaction[SrcChain],
            ) -> Coroutine[Any, Any, OrderResponse]:
                logger.debug("Not using referral code for CCTP orders")
                return self.submit_cctp_order(sent_transaction)

        if isinstance(src_token, ApprovableToken):
            if approval_error := await self.approve_token(
                src_token, account, spender, amount, logger
            ):
                return approval_error

        logger.info(f"Placing {error_data['type']} transaction")

        try:
            transaction = await create_transaction_coro
        except Exception as e:
            logger.error(
                f"Failed to create {error_data['type']} transaction:", exc_info=e
            )
            return CreateTransactionFailed(e, error_data)

        result = await self.place_logged_transaction(transaction, logger)

        if isinstance(result, TransactionError):
            return result

        logger.info("Order placed. Submitting to backend.")

        try:
            order_response = await submit_order(result)
        except Exception as e:
            logger.error("Error submitting order to backend:", exc_info=e)
            return SubmitOrderFailed(
                (src_token, dest_token),
                amount,
                result,
                error_data,
                e,
            )

        logger.info("Order submitted to backend with response:")
        logger.info(pprint.pformat(order_response))
        logger.info("Waiting for order status updates")

        start_time = datetime.now()
        filled_at = (
            datetime.fromisoformat(order_response.filled_at)
            if order_response.filled_at
            else None
        )

        async for order in self.poll_order_status(order_response, 15):
            # Fallback on checking the token balances to determine if the order was filled or completed
            if not order:
                now = datetime.now()
                time_elapsed = now - start_time

                if filled_at:
                    received_amount = (
                        await dest_token.get_balance(recipient) - initial_dest_balance
                    )

                    if received_amount == 0:
                        if now > filled_at + max_wait_time:
                            logger.warning(
                                "Destination funds not received after max waiting time"
                            )
                            return DestinationNotReceived(
                                (src_token, dest_token),
                                amount,
                                order_response,
                                time_elapsed,
                            )

                        continue

                    elif received_amount != quote.dst_amount:
                        logger.warning(
                            f"Expected to receive {quote.dst_amount} {dest_token}, actually received {received_amount}"
                        )
                        assert received_amount > 0

                    logger.info("Order complete but websocket endpoint may be down")
                    break

                else:
                    filled_amount = initial_src_balance - await src_token.get_balance(
                        account.downcast()
                    )

                    if filled_amount == 0:
                        if now > start_time + max_wait_time:
                            logger.warning("Order not filled after max waiting time")
                            return SourceNotWithdrawn(
                                (src_token, dest_token),
                                amount,
                                order_response,
                                time_elapsed,
                            )

                        continue

                    elif filled_amount != amount:
                        logger.warning(
                            f"Expected to fill {amount} {src_token}, actually filled {filled_amount}"
                        )
                        assert filled_amount > 0

                    filled_at = now
                    logger.info("Order filled but websocket may be down")
                    continue

            if order.completed:
                break

            elif order.filled_at and not filled_at:
                filled_at = datetime.fromisoformat(order.filled_at)
                logger.info(f"Order filled at {filled_at}: {order}")

            else:
                logger.debug(f"Received order update: {order}")

        logger.info("Trade complete")

        return Trade((src_token, dest_token), quote, order_response)


__all__ = ["AssetServer", "MachClient", "RiskManager"]



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class SetMaxBurnAmountPerMessageArgs(typing.TypedDict):
    params: types.set_max_burn_amount_per_message_params.SetMaxBurnAmountPerMessageParams


layout = borsh.CStruct(
    "params"
    / types.set_max_burn_amount_per_message_params.SetMaxBurnAmountPerMessageParams.layout
)


class SetMaxBurnAmountPerMessageAccounts(typing.TypedDict):
    token_controller: Pubkey
    token_minter: Pubkey
    local_token: Pubkey
    event_authority: Pubkey
    program: Pubkey


def set_max_burn_amount_per_message(
    args: SetMaxBurnAmountPerMessageArgs,
    accounts: SetMaxBurnAmountPerMessageAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(
            pubkey=accounts["token_controller"], is_signer=True, is_writable=False
        ),
        AccountMeta(
            pubkey=accounts["token_minter"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["local_token"], is_signer=False, is_writable=True),
        AccountMeta(
            pubkey=accounts["event_authority"], is_signer=False, is_writable=False
        ),
        AccountMeta(pubkey=accounts["program"], is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x1e\x80\x91\xf0F\xedm\xcf"
    encoded_args = layout.build(
        {
            "params": args["params"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import dataclasses
import typing

from solana.rpc.async_api import AsyncClient

from .. import config
from ..chain import SolanaChain
from .chain_client import ChainClient


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class SolanaClient[Chain: SolanaChain](ChainClient[Chain]):
    native: AsyncClient

    @classmethod
    @typing.override
    async def create(cls, chain: Chain) -> SolanaClient[Chain]:
        client = AsyncClient(config.config.endpoint_uris[chain])
        return cls(chain=chain, native=client)

    @typing.override
    async def close(self) -> None:
        if await self.is_connected():
            await self.native.close()

    @typing.override
    async def is_connected(self) -> bool:
        return await self.native.is_connected()

    @typing.override
    async def reconnect(self) -> None:
        await super().reconnect()



import typing
from . import register_tristero_o_app_params
from .register_tristero_o_app_params import (
    RegisterTristeroOAppParams,
    RegisterTristeroOAppParamsJSON,
)
from . import update_params
from .update_params import UpdateParams, UpdateParamsJSON
from . import lz_receive_type_params
from .lz_receive_type_params import LzReceiveTypeParams, LzReceiveTypeParamsJSON
from . import lz_receive_params
from .lz_receive_params import LzReceiveParams, LzReceiveParamsJSON
from . import confirm_match_params
from .confirm_match_params import ConfirmMatchParams, ConfirmMatchParamsJSON
from . import create_match_params
from .create_match_params import CreateMatchParams, CreateMatchParamsJSON
from . import execute_match_params
from .execute_match_params import ExecuteMatchParams, ExecuteMatchParamsJSON
from . import fill_swap_params
from .fill_swap_params import FillSwapParams, FillSwapParamsJSON
from . import finish_challenge_params
from .finish_challenge_params import FinishChallengeParams, FinishChallengeParamsJSON
from . import place_order_params
from .place_order_params import PlaceOrderParams, PlaceOrderParamsJSON
from . import challenge_params
from .challenge_params import ChallengeParams, ChallengeParamsJSON



from __future__ import annotations
import typing
from dataclasses import dataclass
from construct import Container
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh


class UpdatePauserParamsJSON(typing.TypedDict):
    new_pauser: str


@dataclass
class UpdatePauserParams:
    layout: typing.ClassVar = borsh.CStruct("new_pauser" / BorshPubkey)
    new_pauser: Pubkey

    @classmethod
    def from_decoded(cls, obj: Container) -> "UpdatePauserParams":
        return cls(new_pauser=obj.new_pauser)

    def to_encodable(self) -> dict[str, typing.Any]:
        return {"new_pauser": self.new_pauser}

    def to_json(self) -> UpdatePauserParamsJSON:
        return {"new_pauser": str(self.new_pauser)}

    @classmethod
    def from_json(cls, obj: UpdatePauserParamsJSON) -> "UpdatePauserParams":
        return cls(new_pauser=Pubkey.from_string(obj["new_pauser"]))



import typing
from . import accept_ownership_params
from .accept_ownership_params import AcceptOwnershipParams, AcceptOwnershipParamsJSON
from . import add_remote_token_messenger_params
from .add_remote_token_messenger_params import (
    AddRemoteTokenMessengerParams,
    AddRemoteTokenMessengerParamsJSON,
)
from . import deposit_for_burn_with_caller_params
from .deposit_for_burn_with_caller_params import (
    DepositForBurnWithCallerParams,
    DepositForBurnWithCallerParamsJSON,
)
from . import deposit_for_burn_params
from .deposit_for_burn_params import DepositForBurnParams, DepositForBurnParamsJSON
from . import handle_receive_message_params
from .handle_receive_message_params import (
    HandleReceiveMessageParams,
    HandleReceiveMessageParamsJSON,
)
from . import initialize_params
from .initialize_params import InitializeParams, InitializeParamsJSON
from . import remove_remote_token_messenger_params
from .remove_remote_token_messenger_params import (
    RemoveRemoteTokenMessengerParams,
    RemoveRemoteTokenMessengerParamsJSON,
)
from . import replace_deposit_for_burn_params
from .replace_deposit_for_burn_params import (
    ReplaceDepositForBurnParams,
    ReplaceDepositForBurnParamsJSON,
)
from . import transfer_ownership_params
from .transfer_ownership_params import (
    TransferOwnershipParams,
    TransferOwnershipParamsJSON,
)
from . import add_local_token_params
from .add_local_token_params import AddLocalTokenParams, AddLocalTokenParamsJSON
from . import burn_token_custody_params
from .burn_token_custody_params import (
    BurnTokenCustodyParams,
    BurnTokenCustodyParamsJSON,
)
from . import link_token_pair_params
from .link_token_pair_params import LinkTokenPairParams, LinkTokenPairParamsJSON
from . import pause_params
from .pause_params import PauseParams, PauseParamsJSON
from . import remove_local_token_params
from .remove_local_token_params import (
    RemoveLocalTokenParams,
    RemoveLocalTokenParamsJSON,
)
from . import set_max_burn_amount_per_message_params
from .set_max_burn_amount_per_message_params import (
    SetMaxBurnAmountPerMessageParams,
    SetMaxBurnAmountPerMessageParamsJSON,
)
from . import set_token_controller_params
from .set_token_controller_params import (
    SetTokenControllerParams,
    SetTokenControllerParamsJSON,
)
from . import unink_token_pair_params
from .unink_token_pair_params import UninkTokenPairParams, UninkTokenPairParamsJSON
from . import unpause_params
from .unpause_params import UnpauseParams, UnpauseParamsJSON
from . import update_pauser_params
from .update_pauser_params import UpdatePauserParams, UpdatePauserParamsJSON
from . import token_minter_error
from .token_minter_error import TokenMinterErrorKind, TokenMinterErrorJSON



from __future__ import annotations
import abc
import dataclasses
from decimal import Decimal
from collections import defaultdict
import typing
from typing import Any, ClassVar, Optional

from ..chain import Chain
from ..chain_client import ChainClient, EthereumClient, SolanaClient, TronClient
from ..account import Account, AccountID
from ..transaction import SentTransaction
from .asset import Asset


NATIVE_COIN_ADDRESS = "native"


# Fungible token
# Caches some frequently accessed metadata
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class Token[ChainType: Chain](Asset[ChainType]):
    lookup_cache: ClassVar[defaultdict[Chain, dict[str, Token]]] = defaultdict(dict)

    symbol: str
    decimals: int

    # The client calls this while initializing the configuration, which caches the tokens so that you can look them up by name
    @classmethod
    async def register(
        cls,
        client: ChainClient[ChainType],
        address: str,
        symbol: Optional[str],
        decimals: Optional[int],
    ) -> Token[ChainType]:
        if token := cls.lookup_cache[client.chain].get(address):
            return token

        elif address == NATIVE_COIN_ADDRESS:
            assert (
                symbol and decimals
            ), "Both symbol and decimals must be provided for the native coin"

            match client:
                case EthereumClient():
                    coin_class = ethereum.EthereumNativeCoin
                case SolanaClient():
                    coin_class = solana.SolanaNativeCoin
                case TronClient():
                    coin_class = tron.TronNativeCoin
                case _:
                    raise NotImplementedError(f"Unimplemented chain: {client.chain}")

            token = coin_class(
                client=client,  # type: ignore
                symbol=symbol,
                decimals=decimals,
            )

        else:
            match client:
                case EthereumClient():
                    token = await ethereum.EthereumToken.from_data(
                        client, address, symbol, decimals
                    )
                case SolanaClient():
                    token = await solana.SolanaToken.from_data(
                        client, address, symbol, decimals
                    )
                case TronClient():
                    token = await tron.TronToken.from_data(
                        client, address, symbol, decimals
                    )
                case _:
                    raise NotImplementedError(f"Unimplemented chain: {client.chain}")

        cls.lookup_cache[token.chain].update(
            ((token.symbol, token), (token.address, token))
        )

        return token

    @classmethod
    def try_lookup(cls, chain: ChainType, key: str) -> Optional[Token[ChainType]]:
        return cls.lookup_cache[chain].get(key)

    @classmethod
    def lookup(cls, chain: ChainType, key: str) -> Token[ChainType]:
        return cls.lookup_cache[chain][key]

    @classmethod
    def lookup_symbol(cls, chain: ChainType, symbol: str) -> Token[ChainType]:
        return cls.lookup(chain, symbol)

    @classmethod
    def try_lookup_symbol(
        cls, chain: ChainType, symbol: str
    ) -> Optional[Token[ChainType]]:
        return cls.try_lookup(chain, symbol)

    @classmethod
    def lookup_address(cls, chain: ChainType, address: str) -> Token[ChainType]:
        return cls.lookup(chain, address)

    @classmethod
    def try_lookup_address(
        cls, chain: ChainType, address: str
    ) -> Optional[Token[ChainType]]:
        return cls.lookup_cache[chain].get(address)

    @classmethod
    def from_str(cls, string: str) -> Token[ChainType]:
        chain, symbol = string.split("-")
        return cls.lookup_symbol(Chain.from_str(chain), symbol)  # type: ignore

    @classmethod
    def from_id_str(cls, id: str) -> Token[ChainType]:
        chain_id, rest = id.split("/")
        chain = Chain.from_id_str(chain_id)
        _asset_namespace, address = rest.split(":")
        return cls.lookup_address(chain, address)  # type: ignore

    @property
    @typing.override
    def asset_reference(self) -> str:
        return self.address

    @property
    @abc.abstractmethod
    def address(self) -> str:
        pass

    @abc.abstractmethod
    def encode_address(self) -> bytes:
        pass

    @abc.abstractmethod
    async def get_balance(self, account_id: AccountID[ChainType]) -> int:
        pass

    async def get_balance_in_coins(self, account_id: AccountID[ChainType]) -> Decimal:
        return Decimal(await self.get_balance(account_id)) / 10**self.decimals

    def to_coins(self, amount: int) -> Decimal:
        return Decimal(amount) / 10**self.decimals

    def format_amount(self, amount: int) -> str:
        return f"{self.to_coins(amount)} {self}"

    @abc.abstractmethod
    async def transfer(
        self,
        sender: Account[ChainType],
        recipient: AccountID[ChainType],
        amount: int,
    ) -> tuple[SentTransaction[ChainType], Any]:
        pass

    async def transfer_checked(
        self,
        sender: Account[ChainType],
        recipient: AccountID[ChainType],
        amount: int,
    ) -> tuple[SentTransaction[ChainType], Any]:
        assert (
            await self.get_balance(sender.downcast()) >= amount
        ), "Insufficient balance"
        return await self.transfer(sender, recipient, amount)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self})"

    def __str__(self) -> str:
        return f"{self.chain}-{self.symbol}"

    def is_native(self) -> bool:
        return self.address == NATIVE_COIN_ADDRESS

    def is_stablecoin(self) -> bool:
        return self.symbol in ("FRAX", "DAI", "MIM") or any(
            map(
                lambda symbol: symbol in self.symbol,
                ("USD", "EUR", "JPY", "GPB", "CHF"),
            )
        )

    def is_chf_stablecoin(self) -> bool:
        return "CHF" in self.symbol

    def is_eur_stablecoin(self) -> bool:
        return "EUR" in self.symbol

    def is_gbp_stablecoin(self) -> bool:
        return "GBP" in self.symbol

    def is_jpy_stablecoin(self) -> bool:
        return "JPY" in self.symbol

    def is_usd_stablecoin(self) -> bool:
        return "USD" in self.symbol or self.symbol in ("FRAX", "DAI", "MIM")

    def is_btc(self) -> bool:
        return "BTC" in self.symbol

    def is_eth(self) -> bool:
        return "ETH" in self.symbol


# The gas token/native asset for a chain
@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class NativeCoin[ChainType: Chain](Token[ChainType]):
    # https://chainagnostic.org/CAIPs/caip-20
    @property
    @typing.override
    def asset_namespace(self) -> str:
        return "slip44"

    # TODO: We might want to raise NotImplementedError here, but we're doing this for compatibility with the asset server
    @property
    @typing.override
    def address(self) -> str:
        return NATIVE_COIN_ADDRESS
    
    @typing.override
    def encode_address(self) -> bytes:
        raise NotImplementedError(f"Address not encodable: {self.address}")


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class ApprovableToken[ChainType: Chain](Token[ChainType]):
    @abc.abstractmethod
    async def get_allowance(
        self,
        owner: AccountID[ChainType],
        spender: AccountID[ChainType],
    ) -> int:
        pass

    @abc.abstractmethod
    async def approve(
        self,
        owner: Account[ChainType],
        spender: AccountID[ChainType],
        amount: int,
    ) -> Any:
        pass


# Avoid a circular import
from . import ethereum, solana, tron  # noqa: E402



import typing
from anchorpy.error import ProgramError


class InvalidAuthority(ProgramError):
    def __init__(self) -> None:
        super().__init__(6000, "Invalid authority")

    code = 6000
    name = "InvalidAuthority"
    msg = "Invalid authority"


class InvalidTokenMessengerState(ProgramError):
    def __init__(self) -> None:
        super().__init__(6001, "Invalid token messenger state")

    code = 6001
    name = "InvalidTokenMessengerState"
    msg = "Invalid token messenger state"


class InvalidTokenMessenger(ProgramError):
    def __init__(self) -> None:
        super().__init__(6002, "Invalid token messenger")

    code = 6002
    name = "InvalidTokenMessenger"
    msg = "Invalid token messenger"


class InvalidOwner(ProgramError):
    def __init__(self) -> None:
        super().__init__(6003, "Invalid owner")

    code = 6003
    name = "InvalidOwner"
    msg = "Invalid owner"


class MalformedMessage(ProgramError):
    def __init__(self) -> None:
        super().__init__(6004, "Malformed message")

    code = 6004
    name = "MalformedMessage"
    msg = "Malformed message"


class InvalidMessageBodyVersion(ProgramError):
    def __init__(self) -> None:
        super().__init__(6005, "Invalid message body version")

    code = 6005
    name = "InvalidMessageBodyVersion"
    msg = "Invalid message body version"


class InvalidAmount(ProgramError):
    def __init__(self) -> None:
        super().__init__(6006, "Invalid amount")

    code = 6006
    name = "InvalidAmount"
    msg = "Invalid amount"


class InvalidDestinationDomain(ProgramError):
    def __init__(self) -> None:
        super().__init__(6007, "Invalid destination domain")

    code = 6007
    name = "InvalidDestinationDomain"
    msg = "Invalid destination domain"


class InvalidDestinationCaller(ProgramError):
    def __init__(self) -> None:
        super().__init__(6008, "Invalid destination caller")

    code = 6008
    name = "InvalidDestinationCaller"
    msg = "Invalid destination caller"


class InvalidMintRecipient(ProgramError):
    def __init__(self) -> None:
        super().__init__(6009, "Invalid mint recipient")

    code = 6009
    name = "InvalidMintRecipient"
    msg = "Invalid mint recipient"


class InvalidSender(ProgramError):
    def __init__(self) -> None:
        super().__init__(6010, "Invalid sender")

    code = 6010
    name = "InvalidSender"
    msg = "Invalid sender"


class InvalidTokenPair(ProgramError):
    def __init__(self) -> None:
        super().__init__(6011, "Invalid token pair")

    code = 6011
    name = "InvalidTokenPair"
    msg = "Invalid token pair"


class InvalidTokenMint(ProgramError):
    def __init__(self) -> None:
        super().__init__(6012, "Invalid token mint")

    code = 6012
    name = "InvalidTokenMint"
    msg = "Invalid token mint"


CustomError = typing.Union[
    InvalidAuthority,
    InvalidTokenMessengerState,
    InvalidTokenMessenger,
    InvalidOwner,
    MalformedMessage,
    InvalidMessageBodyVersion,
    InvalidAmount,
    InvalidDestinationDomain,
    InvalidDestinationCaller,
    InvalidMintRecipient,
    InvalidSender,
    InvalidTokenPair,
    InvalidTokenMint,
]
CUSTOM_ERROR_MAP: dict[int, CustomError] = {
    6000: InvalidAuthority(),
    6001: InvalidTokenMessengerState(),
    6002: InvalidTokenMessenger(),
    6003: InvalidOwner(),
    6004: MalformedMessage(),
    6005: InvalidMessageBodyVersion(),
    6006: InvalidAmount(),
    6007: InvalidDestinationDomain(),
    6008: InvalidDestinationCaller(),
    6009: InvalidMintRecipient(),
    6010: InvalidSender(),
    6011: InvalidTokenPair(),
    6012: InvalidTokenMint(),
}


def from_code(code: int) -> typing.Optional[CustomError]:
    maybe_err = CUSTOM_ERROR_MAP.get(code)
    if maybe_err is None:
        return None
    return maybe_err



import typing

from eth_typing import ChecksumAddress

from ... import config
from ...account import EthereumAccount
from ...chain import EthereumChain
from ...chain_client import EthereumClient
from ...transaction import EthereumTransaction
from ..types import OrderData


async def create_cctp_burn_transaction[SrcChainType: EthereumChain](
    *,
    src_client: EthereumClient[SrcChainType],
    destination_domain: int,
    token_messenger_minter_address: str,
    order_data: OrderData,
    account: EthereumAccount[SrcChainType],
    recipient: bytes,
) -> EthereumTransaction[SrcChainType]:
    contract = src_client.w3.eth.contract(
        address=typing.cast(ChecksumAddress, token_messenger_minter_address),
        abi=config.ethereum_cctp_token_messenger_abi,
    )

    burn_function = contract.functions.depositForBurn(
        order_data.order_funding.src_amount_in,
        destination_domain,
        recipient,
        order_data.order_direction.src_token_address,
    )

    return await EthereumTransaction.from_contract_function(
        src_client, burn_function, account
    )



from .chain import Chain
from .constants import SupportedChain
from .ethereum import EthereumChain
from .solana import SolanaChain
from .tron import TronChain


__all__ = [
    "Chain",
    "EthereumChain",
    "SolanaChain",
    "SupportedChain",
    "TronChain",
]



from __future__ import annotations
import abc
from abc import ABC
import dataclasses
import typing
from typing import Any

from ..chain import Chain


# CAIP-19 Asset Type
# https://chainagnostic.org/CAIPs/caip-19
@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class Asset[ChainType: Chain](ABC):
    @property
    @abc.abstractmethod
    def chain(self) -> ChainType:
        pass

    @property
    @abc.abstractmethod
    def asset_namespace(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def asset_reference(self) -> str:
        pass

    @property
    def id(self) -> str:
        return f"{self.chain.id}/{self.asset_namespace}:{self.asset_reference}"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Asset) and self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id})"

    def __str__(self) -> str:
        return f"{self.chain}-{self.asset_reference}"


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class GenericAsset[ChainType: Chain](Asset[ChainType]):
    chain_: ChainType
    asset_namespace_: str
    asset_reference_: str

    @classmethod
    def from_id_str(cls, id: str) -> GenericAsset[ChainType]:
        chain_id, rest = id.split("/")
        chain = Chain.from_id_str(chain_id)
        asset_namespace, asset_reference = rest.split(":")
        return cls(
            chain_=chain,  # type: ignore
            asset_namespace_=asset_namespace,
            asset_reference_=asset_reference,
        )

    @property
    @typing.override
    def chain(self) -> ChainType:
        return self.chain_

    @property
    @typing.override
    def asset_namespace(self) -> str:
        return self.asset_namespace_

    @property
    @typing.override
    def asset_reference(self) -> str:
        return self.asset_reference_



from __future__ import annotations
import dataclasses
import typing
from typing import Optional

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.contract import AsyncContract
from web3.types import TxParams, TxReceipt

from .. import config
from ..account import EthereumAccount, EthereumAccountID
from ..chain import EthereumChain
from ..chain_client import EthereumClient
from ..utility import ethereum
from ..transaction import EthereumSentTransaction, EthereumTransaction
from .token import ApprovableToken, NativeCoin


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class EthereumNativeCoin[Chain: EthereumChain](NativeCoin[Chain]):
    client: EthereumClient[Chain]

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    @typing.override
    async def get_balance(self, account_id: EthereumAccountID[Chain]) -> int:
        return await self.client.w3.eth.get_balance(account_id.address)

    @typing.override
    async def transfer(
        self,
        sender: EthereumAccount[Chain],
        recipient: EthereumAccountID[Chain],
        amount: int,
    ) -> tuple[EthereumSentTransaction[Chain], TxReceipt]:
        params = TxParams(
            {
                "from": sender.native.address,
                "to": recipient.address_,
                "value": self.client.w3.to_wei(amount, "wei"),
            }
        )

        transaction = await EthereumTransaction.from_transaction_params(
            self.client, params, sender
        )

        sent_transaction = await transaction.broadcast()
        return sent_transaction, await sent_transaction.wait_for_receipt()


@dataclasses.dataclass(frozen=True, kw_only=True, repr=False, slots=True)
class EthereumToken[Chain: EthereumChain](ApprovableToken[Chain]):
    client: EthereumClient[Chain]
    contract: AsyncContract

    @classmethod
    async def from_data(
        cls,
        client: EthereumClient[Chain],
        address: str,
        symbol: Optional[str],
        decimals: Optional[int],
    ) -> EthereumToken[Chain]:
        contract = client.w3.eth.contract(
            address=client.w3.to_checksum_address(address),
            abi=config.erc20_abi,
        )

        if not symbol:
            symbol = typing.cast(str, await contract.functions.symbol().call())

        if not decimals:
            decimals = typing.cast(int, await contract.functions.decimals().call())

        return cls(client=client, contract=contract, symbol=symbol, decimals=decimals)

    @property
    @typing.override
    def chain(self) -> Chain:
        return self.client.chain

    # https://namespaces.chainagnostic.org/eip155/caip19
    @property
    @typing.override
    def asset_namespace(self) -> str:
        return "erc20"

    @property
    @typing.override
    def address(self) -> ChecksumAddress:
        return self.contract.address

    @typing.override
    def encode_address(self) -> bytes:
        return ethereum.encode_address(self.contract.address)

    @typing.override
    async def get_balance(self, account_id: EthereumAccountID[Chain]) -> int:
        return await self.contract.functions.balanceOf(account_id.address).call()

    async def create_transfer_transaction(
        self,
        sender: EthereumAccount[Chain],
        recipient: EthereumAccountID[Chain],
        amount: int,
    ) -> EthereumTransaction[Chain]:
        contract_function = self.contract.functions.transfer(recipient.address, amount)
        return await EthereumTransaction.from_contract_function(
            self.client, contract_function, sender
        )

    @typing.override
    async def transfer(
        self,
        sender: EthereumAccount[Chain],
        recipient: EthereumAccountID[Chain],
        amount: int,
    ) -> tuple[EthereumSentTransaction[Chain], TxReceipt]:
        transaction = await self.create_transfer_transaction(sender, recipient, amount)
        sent_transaction = await transaction.broadcast()
        return sent_transaction, await sent_transaction.wait_for_receipt()

    @typing.override
    async def get_allowance(
        self,
        owner: EthereumAccountID[Chain],
        spender: EthereumAccountID[Chain],
    ) -> int:
        return await self.contract.functions.allowance(
            owner.address, spender.address
        ).call()

    async def create_approval_transaction(
        self,
        owner: EthereumAccount[Chain],
        spender: EthereumAccountID[Chain],
        amount: int,
    ) -> EthereumTransaction[Chain]:
        contract_function = self.contract.functions.approve(spender.address, amount)
        return await EthereumTransaction.from_contract_function(
            self.client, contract_function, owner
        )

    @typing.override
    async def approve(
        self,
        owner: EthereumAccount[Chain],
        spender: EthereumAccountID[Chain],
        amount: int,
    ) -> TxReceipt:
        transaction = await self.create_approval_transaction(owner, spender, amount)
        sent_transaction = await transaction.broadcast()
        return await sent_transaction.wait_for_receipt()



from __future__ import annotations
import dataclasses
import typing

from tronpy.async_tron import AsyncTron
from tronpy.providers import AsyncHTTPProvider

from .. import config
from ..chain import TronChain
from .chain_client import ChainClient


@dataclasses.dataclass(frozen=True, kw_only=True, slots=True)
class TronClient[Chain: TronChain](ChainClient[Chain]):
    native: AsyncTron

    @classmethod
    @typing.override
    async def create(cls, chain: Chain) -> TronClient[Chain]:
        provider = AsyncHTTPProvider(config.config.endpoint_uris[chain], timeout=60.0)
        client = AsyncTron(provider)
        return cls(chain=chain, native=client)

    @typing.override
    async def close(self) -> None:
        if await self.is_connected():
            await self.native.close()

    @typing.override
    async def is_connected(self) -> bool:
        # TODO: is_closed tells us if the connection is closed, but we should also check if the connection is open
        # However, the state enum that tells us if the connection is open is not exposed by the library
        return not self.native.provider.client.is_closed

    @typing.override
    async def reconnect(self) -> None:
        await super().reconnect()


from __future__ import annotations
import typing

from ...account import (
    Account,
    AccountID,
    EthereumAccount,
    SolanaAccount,
    SolanaAccountID,
)
from ...asset import SolanaToken, Token
from ...chain import Chain
from ...chain_client import (
    ChainClient,
    EthereumClient,
    SolanaClient,
)
from ...transaction import Transaction
from ..types import OrderData
from . import ethereum, solana

if typing.TYPE_CHECKING:
    from ...client import MachClient


async def create_cctp_burn_transaction[SrcChainType: Chain, DestChainType: Chain](
    *,
    client: MachClient,
    src_client: ChainClient[SrcChainType],
    src_token: Token[SrcChainType],
    dest_token: Token[DestChainType],
    order_data: OrderData,
    account: Account[SrcChainType],
    recipient: AccountID[DestChainType],
) -> Transaction[SrcChainType]:
    token_messenger_minter_address = client.cctp_token_messenger_minter_address(
        src_client.chain
    )
    destination_domain = recipient.chain.cctp_domain

    # See the [Mint recipient for Solana as Destination Chain section](https://developers.circle.com/stablecoins/solana-programs)
    if isinstance(recipient, SolanaAccountID):
        recipient_token_account = typing.cast(
            SolanaToken, dest_token
        ).associated_token_account(recipient.pubkey)
        recipient_bytes = bytes(recipient_token_account)
    else:
        recipient_bytes = recipient.encode_address()

    recipient_bytes = recipient_bytes.rjust(32, b"\0")

    match src_client:
        case EthereumClient():
            return await ethereum.create_cctp_burn_transaction(
                src_client=src_client,
                destination_domain=destination_domain,
                token_messenger_minter_address=token_messenger_minter_address,
                order_data=order_data,
                account=typing.cast(EthereumAccount, account),
                recipient=recipient_bytes,
            )

        case SolanaClient():
            return await solana.create_cctp_burn_transaction(
                client=client,
                src_client=src_client,
                src_token=typing.cast(SolanaToken, src_token),
                destination_domain=destination_domain,
                token_messenger_minter_address=token_messenger_minter_address,
                order_data=order_data,
                account=typing.cast(SolanaAccount, account),
                recipient=recipient_bytes,
            )

        case _:
            raise NotImplementedError(src_client.chain)


__all__ = ["create_cctp_burn_transaction"]



import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment
import borsh_construct as borsh
from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.error import AccountInvalidDiscriminator
from anchorpy.utils.rpc import get_multiple_accounts
from anchorpy.borsh_extension import BorshPubkey
from ..program_id import PROGRAM_ID


class ReceiptJSON(typing.TypedDict):
    maker: str
    payout_quantity: int
    token_mint: str
    receiver: str
    is_valuable: bool


@dataclass
class Receipt:
    discriminator: typing.ClassVar = b"'\x9aIjPf\x91\x99"
    layout: typing.ClassVar = borsh.CStruct(
        "maker" / BorshPubkey,
        "payout_quantity" / borsh.U64,
        "token_mint" / BorshPubkey,
        "receiver" / BorshPubkey,
        "is_valuable" / borsh.Bool,
    )
    maker: Pubkey
    payout_quantity: int
    token_mint: Pubkey
    receiver: Pubkey
    is_valuable: bool

    @classmethod
    async def fetch(
        cls,
        conn: AsyncClient,
        address: Pubkey,
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.Optional["Receipt"]:
        resp = await conn.get_account_info(address, commitment=commitment)
        info = resp.value
        if info is None:
            return None
        if info.owner != program_id:
            raise ValueError("Account does not belong to this program")
        bytes_data = info.data
        return cls.decode(bytes_data)

    @classmethod
    async def fetch_multiple(
        cls,
        conn: AsyncClient,
        addresses: list[Pubkey],
        commitment: typing.Optional[Commitment] = None,
        program_id: Pubkey = PROGRAM_ID,
    ) -> typing.List[typing.Optional["Receipt"]]:
        infos = await get_multiple_accounts(conn, addresses, commitment=commitment)
        res: typing.List[typing.Optional["Receipt"]] = []
        for info in infos:
            if info is None:
                res.append(None)
                continue
            if info.account.owner != program_id:
                raise ValueError("Account does not belong to this program")
            res.append(cls.decode(info.account.data))
        return res

    @classmethod
    def decode(cls, data: bytes) -> "Receipt":
        if data[:ACCOUNT_DISCRIMINATOR_SIZE] != cls.discriminator:
            raise AccountInvalidDiscriminator(
                "The discriminator for this account is invalid"
            )
        dec = Receipt.layout.parse(data[ACCOUNT_DISCRIMINATOR_SIZE:])
        return cls(
            maker=dec.maker,
            payout_quantity=dec.payout_quantity,
            token_mint=dec.token_mint,
            receiver=dec.receiver,
            is_valuable=dec.is_valuable,
        )

    def to_json(self) -> ReceiptJSON:
        return {
            "maker": str(self.maker),
            "payout_quantity": self.payout_quantity,
            "token_mint": str(self.token_mint),
            "receiver": str(self.receiver),
            "is_valuable": self.is_valuable,
        }

    @classmethod
    def from_json(cls, obj: ReceiptJSON) -> "Receipt":
        return cls(
            maker=Pubkey.from_string(obj["maker"]),
            payout_quantity=obj["payout_quantity"],
            token_mint=Pubkey.from_string(obj["token_mint"]),
            receiver=Pubkey.from_string(obj["receiver"]),
            is_valuable=obj["is_valuable"],
        )



from __future__ import annotations
import asyncio
from collections import defaultdict
import dataclasses
from decimal import Decimal
import heapq
import itertools
import logging
from typing import Optional, Sequence

from aiohttp import ClientSession
import cachebox
from cachebox import Cache, TTLCache
from pydantic import TypeAdapter

from ... import config
from ...account import AccountBase, AccountIDManager
from ...asset import Token
from ...chain import Chain, SupportedChain
from ...chain_client import ChainClient
from ...log import LogContextAdapter, Logger
from .. import utility
from ..types import MachChain
from .types import AssetInfo, AssetPricingData, UserAssetData


# This is a completely separate client instead of a mixin like the miles server because it has to be initialized, which is expensive and unnecessary for some use cases
class AssetServer:
    @dataclasses.dataclass(kw_only=True, slots=True)
    class CoinInfo:
        asset_info: AssetInfo
        tokens: set[Token]

    __slots__ = (
        "logger",
        "routes",
        "session",
        "asset_info_heap",
        "coin_info_map",
        "token_info_map",
        "tokens",
    )

    @classmethod
    @cachebox.cached(Cache(0), copy_level=2)
    async def create(
        cls,
        *,
        logger: Logger = logging.getLogger("mach-client"),
    ) -> AssetServer:
        client = cls(logger)
        await client.refresh_assets()
        return client

    def __init__(self, logger: Logger) -> None:
        self.logger = LogContextAdapter(logger, "Asset Server")

        self.routes = config.config.token_server.endpoints.add_url(
            config.config.token_server.url
        )
        self.session = ClientSession()
        self.session.headers.update(
            (
                ("accept", "application/json"),
                ("Content-Type", "application/json"),
            )
        )

        # Heap of AssetInfos ordered by market cap rank
        self.asset_info_heap: list[AssetInfo] = []
        # ID | symbol -> CoinInfo
        self.coin_info_map: dict[str, AssetServer.CoinInfo] = {}
        # Token -> AssetInfo
        self.token_info_map: dict[Token, AssetInfo] = {}
        # Chain -> set[Token]
        self.tokens: defaultdict[Chain, set[Token]] = defaultdict(set)

    async def close(self) -> None:
        await self.session.close()

    ### MARKET CAP ###

    # Top coins by market cap
    def get_top_coins(self, limit: int) -> Sequence[AssetInfo]:
        # Note: This data is possibly stale. Only rely on it for name/symbol info.
        return heapq.nsmallest(limit, self.asset_info_heap)

    ### INFO ###

    def is_supported(self, token: Token) -> bool:
        return token in self.tokens[token.chain]

    # This is to allow the following lookups:
    # 0. CoinGecko ID | Symbol -> CoinInfo (other direction is built into CoinInfo.AssetInfo)
    # 1. Token -> AssetInfo
    # 2. Coins in market cap order
    async def _register_asset[ChainType: Chain](
        self,
        *,
        client: ChainClient[ChainType],
        address: str,
        asset_info: AssetInfo,
    ) -> Token[ChainType]:
        token = await Token.register(
            client=client,
            address=address,
            symbol=asset_info.symbol,
            decimals=asset_info.decimals,
        )

        if asset_info.id not in self.coin_info_map:
            coin_info = AssetServer.CoinInfo(asset_info=asset_info, tokens={token})
            self.coin_info_map[asset_info.id] = coin_info
            self.coin_info_map[asset_info.symbol] = coin_info
            heapq.heappush(self.asset_info_heap, asset_info)
        # TODO: Asset info heap is never updated since it would be O(n)
        else:
            coin_info = self.coin_info_map[asset_info.id]
            # Update the asset info
            coin_info.asset_info = asset_info
            coin_info.tokens.add(token)

        self.token_info_map[token] = asset_info
        self.tokens[token.chain].add(token)

        return token

    async def _process_chain_data(
        self, chain: Chain, chain_data: dict[str, AssetInfo]
    ) -> None:
        chain_client = await ChainClient.create(chain)

        self.tokens[chain].update(
            await asyncio.gather(
                *[
                    self._register_asset(
                        client=chain_client,
                        address=address,
                        asset_info=asset_info,
                    )
                    for address, asset_info in chain_data.items()
                ]
            )
        )

    _assets_validator = TypeAdapter(dict[MachChain, dict[str, AssetInfo]])

    # This registers all the tokens returned by the GET /assets endpoint with the client
    async def refresh_assets(self) -> None:
        async with self.session.get(self.routes.assets) as response:
            bytes_result = await utility.to_bytes(response)

        assets = self._assets_validator.validate_json(bytes_result)

        await asyncio.gather(
            *[
                self._process_chain_data(chain.to_chain(), chain_data)
                for chain, chain_data in assets.items()
            ]
        )

    _asset_info_validator = TypeAdapter(Optional[AssetInfo])

    async def _get_asset_info(self, chain: Chain, address: str) -> Optional[AssetInfo]:
        mach_chain = MachChain.from_chain(chain)
        url = f"{self.routes.assets}/{mach_chain.value}/{address}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return self._asset_info_validator.validate_json(bytes_result)

    # Call this if the token in question hasn't yet been registered
    # This gets the asset info and registers the token
    async def get_asset_info(
        self, chain: Chain, address: str
    ) -> Optional[tuple[Token, AssetInfo]]:
        if not (asset_info := await self._get_asset_info(chain, address)):
            return None

        # The refresh_assets method won't register every possible token that you could get info for (ie. it omits very obscure ones)
        # So it's possible we've never seen this token before, and thus should register it first so that it can be looked up
        token = await self._register_asset(
            client=await ChainClient.create(chain),
            address=address,
            asset_info=asset_info,
        )

        return (token, asset_info)

    # Call this if the token in question is already in the `Token.lookup_cache`
    # Note: This means it won't update the asset info cache, you may or may not want to still call `get_asset_info` instead
    async def get_token_info(self, token: Token) -> AssetInfo:
        # The logic is that if you were able to construct a Token object, then it must be in the lookup cache
        # If it was in the lookup cache (which was populated by the asset server), then it must have an asset info
        return await self._get_asset_info(token.chain, token.address)  # type: ignore

    # Note: No guarantees of freshness
    async def get_cached_token_info(self, token: Token) -> AssetInfo:
        if asset_info := self.token_info_map.get(token):
            return asset_info
        return await self.get_token_info(token)

    ### PRICES ###

    async def get_pricing_data(self, token: Token) -> AssetPricingData:
        mach_chain = MachChain.from_chain(token.chain)
        url = f"{self.routes.prices}/{mach_chain.value}/{token.address}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return AssetPricingData.model_validate_json(bytes_result)

    @cachebox.cachedmethod(TTLCache(0, 300.0))
    async def get_cached_price(self, token: Token) -> Decimal:
        pricing_data = await self.get_pricing_data(token)
        return pricing_data.price

    async def get_price(self, token: Token) -> Decimal:
        self.get_cached_price.cache.pop(token)  # type: ignore
        return await self.get_cached_price(token)

    ### BALANCES ###

    _asset_data_validator = TypeAdapter(dict[MachChain, list[UserAssetData]])

    async def get_raw_token_balances(
        self, account: AccountBase
    ) -> dict[MachChain, list[UserAssetData]]:
        url = f"{self.routes.users}/{account.address}/assets"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return self._asset_data_validator.validate_json(bytes_result)

    _chain_asset_data_validator = TypeAdapter(list[UserAssetData])

    async def get_raw_chain_token_balances(
        self, account: AccountBase
    ) -> list[UserAssetData]:
        mach_chain = MachChain.from_chain(account.chain)
        url = f"{self.routes.users}/{account.address}/assets/{mach_chain.value}"

        async with self.session.get(url) as response:
            bytes_result = await utility.to_bytes(response)

        return self._chain_asset_data_validator.validate_json(bytes_result)

    async def _process_chain_balance_data(
        self, chain: Chain, balance_data: list[UserAssetData]
    ) -> dict[Token, int]:
        chain_client = await ChainClient.create(chain)

        tokens = await asyncio.gather(
            *[
                Token.register(
                    client=chain_client,
                    address=asset_data.address,
                    symbol=asset_data.symbol,
                    decimals=None,
                )
                for asset_data in balance_data
            ]
        )

        balances = dict(zip(tokens, map(lambda data: data.balance, balance_data)))
        return balances

    async def get_token_balances(
        self, account: AccountBase
    ) -> dict[Chain, dict[Token, int]]:
        raw_balances = await self.get_raw_token_balances(account)

        account_chain = type(account.chain)

        filtered_balances = [
            (chain, balances)
            for chain, balances in map(
                lambda item: (item[0].to_chain(), item[1]),
                raw_balances.items(),
            )
            if isinstance(chain, account_chain)
        ]

        chains = map(lambda item: item[0], filtered_balances)

        result = await asyncio.gather(
            *[
                self._process_chain_balance_data(chain, balance_data)
                for chain, balance_data in filtered_balances
            ]
        )

        return dict(zip(chains, result))

    async def get_token_balances_in_coins(
        self, account_id: AccountBase
    ) -> dict[Chain, dict[Token, Decimal]]:
        raw_balances = await self.get_token_balances(account_id)
        return utility.balances_in_coins(raw_balances)

    async def get_chain_token_balances(self, account: AccountBase) -> dict[Token, int]:
        raw_balances = await self.get_raw_chain_token_balances(account)
        return await self._process_chain_balance_data(account.chain, raw_balances)

    async def get_chain_token_balances_in_coins(
        self, account: AccountBase
    ) -> dict[Token, Decimal]:
        raw_balances = await self.get_chain_token_balances(account)
        return utility.chain_balances_in_coins(raw_balances)

    async def get_all_raw_token_balances(
        self, accounts: AccountIDManager
    ) -> dict[Chain, list[UserAssetData]]:
        all_accounts = [
            accounts[chain]
            for chain in (
                SupportedChain.ETHEREUM.value,
                SupportedChain.SOLANA.value,
                SupportedChain.TRON.value,
            )
            if chain in accounts
        ]

        balances = await asyncio.gather(
            *[self.get_raw_token_balances(account) for account in all_accounts]
        )

        result: dict[Chain, list[UserAssetData]] = {}

        for account, balances in zip(all_accounts, balances):
            result.update(
                (
                    (chain.to_chain(), chain_balance)
                    for chain, chain_balance in balances.items()
                    if isinstance(chain.to_chain(), type(account.chain))
                )
            )

        return result

    async def get_all_token_balances(
        self, accounts: AccountIDManager
    ) -> dict[Chain, dict[Token, int]]:
        balances = await asyncio.gather(
            *[
                self.get_token_balances(accounts[chain])
                for chain in (
                    SupportedChain.ETHEREUM.value,
                    SupportedChain.SOLANA.value,
                    SupportedChain.TRON.value,
                )
                if chain in accounts
            ]
        )

        return dict(
            itertools.chain(*[chain_balances.items() for chain_balances in balances])
        )

    async def get_all_token_balances_in_coins(
        self, accounts: AccountIDManager
    ) -> dict[Chain, dict[Token, Decimal]]:
        balances = await self.get_all_token_balances(accounts)
        return utility.balances_in_coins(balances)


__all__ = ["AssetServer"]
